/**********************************************************
* Control Moment Gyroscope test script
*
* Created : 19 April 2019
* Author  : Glenn Turner / Gyroscope.com
*
* Notes   : Stepper motor datasheet is here https://www.pololu.com/file/0J687/SY28STH45-0674A.pdf
*           Step angle is 1.8 degrees. e.g. 200 steps per revolution (without microstepping)
*           
*           Stepper motor driver is DRV8834-2 It can handle 3.3 V and 5 V input logic
*           Power for motors should be 2.5v to 10.8 volts
*           Stepper motors are rated for 4.5V at 0.67 A/Phase
*           Design for a Arduino Uno
*           
* Updates : 17/5/2019 Added CalibrateJoystick          
*           23/5/2019 Fixed a number of bugs regarding stepper position
*           16/8/2019 Better start up procedure
*           
*  ARDUINO UNO         
*  DIGITAL PINS
*  0  --> NA
*  1  -->       
*  2  --> STEPPER 4 PULSE               (orange wire)
*  3  --> STEPPER 3 DIR                 (Blue wire)
*  4  --> STEPPER 3 PULSE               (Green wire)
*  5  --> PWM FOR BLDC MOTOR SPEED 1    (Gray wire)
*  6  --> PWM FOR BLDC MOTOR SPEED 2    (Purple wire)     
*  7  --> STEPPER 4 DIR                 (Yellow wire)
*  8  --> STEPPER 2 DIR                 (Gray wire)
*  9  --> STEPPER 2 PULSE               (white wire)    
*  10 --> PWM FOR BLDC MOTOR SPEED 3    (Blue wire)    
*  11 --> PWM FOR BLDC MOTOR SPEED 4    (Green wire)    
*  12 --> STEPPER 1 DIR                 (Black wire)
*  13 --> STEPPER 1 PULSE               (Brown wire)
*  
* GND --> STEPPER MOTOR GND
*  5V --> JOYTSTICK POWER
*  3V --> STEPPER MOTOR CONTROL VOLTAGE REF
* GND --> BLDC GND
*  
*  ANALOGUE PINS
*  0  --> JOYSTICK PITCH
*  1  --> JOYSTICK ROLL
*  2  --> JOYSTICK YAW
*  3  --> JOYSTICK not used
*  4  --> NA
*  5  --> NA
*           
*           
* Copyright : Copyleft. Use as you wish but reference to            
*             Gyroscope.com
**********************************************************/

#include "AccelStepper.h" // Get from https://www.airspayce.com/mikem/arduino/AccelStepper/
#include "filter.h"

#define debugging              true
#define serialbps              115200 // 9600 is another option
#define degperstep             2.2
#define refreshms              1      // How  often we read the joytstick and 
                                      // pulse move the stepper motors one step
#define displayms              400    // How often we update the display (if debugging)
#define volts24                true   // Select if using 20volt or 24 volt supply
#define betterstartingmethod   true   // Do we want the better startup method.
#define steppermaxacc          300    // 600
#define steppermaxspeed        3000   // 3000
#define allowdeviation         7      // deviation allowed from centre (stops jitter on centre)
#define readXtimes             5      // How many times to read joystick values and average

int pitchcentre              = 163;
int rollcentre               = 163;
int yawcentre                = 205;
                              
int pitch = 0;
int roll  = 0;
int yaw   = 0;

long PitchPre = 0;             // previous positions for stepper motors
long RollPre  = 0;
long YawPre   = 0;

long PitchPos = 0;             // Current positions for stepper motors
long RollPos  = 0;
long YawPos   = 0;

long PitchTarget = 0;          // Target positions for stepper motors
long RollTarget  = 0; 
long YawTarget   = 0; 

long displaycounter = 0;

Filter RollFilter(0.03);       // Creates filters for the joystick.
Filter PitchFilter(0.03);      // 0.03 is the filter factor
Filter YawFilter(0.03);        // This is quite a strong filter. 
                               // Removes noise but slows joystick response time


  // Define a steppers and the pins it will use
  AccelStepper stepper1(AccelStepper::DRIVER, 13, 12); // pin 2 is direction, pin 3 is the step
  AccelStepper stepper2(AccelStepper::DRIVER, 9, 8);
  AccelStepper stepper3(AccelStepper::DRIVER, 4, 3);
  AccelStepper stepper4(AccelStepper::DRIVER, 2, 7);

//*********************************************************

void setup() {
  Serial.begin(serialbps);

  stepper1.setMaxSpeed(steppermaxspeed);
  stepper1.setAcceleration(steppermaxacc);

  stepper2.setMaxSpeed(steppermaxspeed);
  stepper2.setAcceleration(steppermaxacc);

  stepper3.setMaxSpeed(steppermaxspeed);
  stepper3.setAcceleration(steppermaxacc); 

  stepper4.setMaxSpeed(steppermaxspeed);
  stepper4.setAcceleration(steppermaxacc);    
 
  pinMode(2, OUTPUT);
  pinMode(3, OUTPUT);
  pinMode(4, OUTPUT);  
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(8, OUTPUT);
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);
  pinMode(11, OUTPUT);
  pinMode(12, OUTPUT);
  pinMode(13, OUTPUT);

  CalibrateJoystick(pitchcentre, rollcentre, yawcentre);
  
  stepper1.setCurrentPosition(0);
  stepper2.setCurrentPosition(0);
  stepper3.setCurrentPosition(0);
  stepper4.setCurrentPosition(0);
  
  delay(3000);

  // Start motors one at a time.
  // Keep amperage spikes and voltage drops to a minimum 

  startBLDCmotor(5, 1);  // BLDC Motor 1
  startBLDCmotor(6, 2);  // BLDC Motor 2
  
  startBLDCmotor(10, 3); // BLDC Motor 3 
  startBLDCmotor(11, 4); // BLDC Motor 4 
}

//*********************************************************

void delayXs(int secs) {
  int j = 0;
  for(j = 1; j <= secs; j++)
  {
    Serial.print(".");
    delay(1000);
  }
}

//*********************************************************

void startBLDCmotor(int thepin, int Motornumber) {
 
  // BLDC motors with no hall sensor struggle to start with heavy mass
  // Controller will fail to start motor unless procedure is correct.
  // The issue is made worse by off-the-shelf controllers (like we are using) 
  // having software timeouts, that assume a light mass is attached.
  // e.g. if the rpm keeps rising after x many seconds. The controller stops the motor.
  //
  // We work around this by.....
  // Getting it spinning. idle the motor, then apply plenty of power and then set
  // a sensible cruise speed.

  if (volts24 == true)
  {
    // This works at 24 volts. won't work at other voltages
    // 0.5 amp per motor at cruise.
    // 5 amps at startup (peak 10 amps)
    // Cruise speed 11045rpm~
    if (betterstartingmethod == true)
    {
      Serial.print("Motor ");
      Serial.print(Motornumber);
      Serial.print(" Just get it spinning");
      analogWrite(thepin, 170); 
      delayXs(4);
      
      Serial.print("Idle motor");
      analogWrite(thepin, 0); 
      delayXs(2);
      
      Serial.print("Starting");
      analogWrite(thepin, 170); 
      delay(500);  //Extra half a second.
      delayXs(5);
      
      Serial.print("Set to a cruise");
      analogWrite(thepin, 110); 
      delayXs(20);
  
      Serial.print("Slow slightly");
      analogWrite(thepin, 105); 
      delayXs(5);
      
      Serial.println("Done.");
    }
    else
    {
      Serial.print("Motor ");
      Serial.print(Motornumber);
      Serial.print(" Just get it spinning");
      analogWrite(thepin, 170); //was 170
      delayXs(4);
      
      Serial.print("Idle motor");
      analogWrite(thepin, 0); //was 170
      delayXs(2);
      
      Serial.print("Starting");
      analogWrite(thepin, 170); //was 170
      delay(500);  //Extra half a second.
      delayXs(5);
      
      Serial.print("Set to a cruise");
      analogWrite(thepin, 110); //153
      delayXs(8);
      
      Serial.println("Done."); 
    }
  }
  else
  {
    // This works at 20 volts. won't work at other voltages
    Serial.println(" Just get it spinning...");
    analogWrite(thepin, 170); //was 170
    delayXs(8);
  
    Serial.println(" Idle motor...");
    analogWrite(thepin, 0); //was 170
    delayXs(3);
    
    Serial.println(" Starting motor...");
    analogWrite(thepin, 170); //was 170
    delayXs(12);
    
    Serial.println(" Set speed to a cruise speed...");
    analogWrite(thepin, 125); //153
    delayXs(5);
  }
}


//*********************************************************

void maxvalue(int& temp) {
  if (temp > 10)
  {
    temp = 10;
  }
  if (temp < -10)
  {
    temp = -10;
  }  
}

//*********************************************************

void CalibrateJoystick(int &pitchcentre, int &rollcentre, int &yawcentre)
{ 
  long pitchtemp = 0;
  long rolltemp = 0;
  long yawtemp = 0;

  Serial.print("Calibrating joystick...");

  PitchFilter.setValue( analogRead(A0));
  RollFilter.setValue(  analogRead(A1));
  YawFilter.setValue(   analogRead(A2));

  for (int i = 0; i < 100;i++ )
  {
    PitchFilter.TEMA( analogRead(A0));
    RollFilter.TEMA(  analogRead(A1));
    YawFilter.TEMA(   analogRead(A2));
    delay(8); // Short to do it over sensible period of time.
  }

  pitchcentre = round(PitchFilter.getValue());
  rollcentre  = round(RollFilter.getValue());
  yawcentre   = round(YawFilter.getValue());

  Serial.println("Done.");

  if (debugging)
  {
    Serial.print(" Joystick Centre Pitch : "); Serial.println(pitchcentre);
    Serial.print(" Joystick Centre Roll  : "); Serial.println(rollcentre);
    Serial.print(" Joystick Centre Yaw   : "); Serial.println(yawcentre);    
  }
}

//*********************************************************

void ReadJoystick(int &pitch, int &roll, int &law)
{
  // Filter values

  int difference = 0;

  pitch =  round(PitchFilter.TEMA( analogRead(A0)));
  roll  =  round(RollFilter.TEMA(analogRead(A1)));
  yaw   =  round(YawFilter.TEMA(analogRead(A2)));

  //ignore the very centre values.
  difference = pitch - pitchcentre;
  difference = abs(difference); 
  if(difference <= allowdeviation)
  {
    pitch = pitchcentre;
  }

  difference = roll - rollcentre;
  difference = abs(difference); 
  if(difference <= allowdeviation)
  {
    roll = rollcentre;
  }

  // Allow for more deviation on yaw axis
  difference = yaw - yawcentre;

  difference = abs(difference);
  if(difference <= allowdeviation*3)
  {
    yaw = yawcentre;
  } 
}

//*********************************************************

bool JoystickPositionChange(int pitch, int roll, int yaw, int pitchpre, int rollpre, int yawpre)
{
  // Joystick position changed?
  if (pitch != pitchpre || yaw != yawpre || roll != rollpre)
  {
    PitchPre = pitch;
    RollPre  = roll;
    YawPre   = yaw;
    return true;
  }
  else
  {
    return false;
  }
}

//*********************************************************

void Joystick2StepperTargets(int thepitch, int theroll, int theyaw, long &PitchTarget, long &RollTarget, long &YawTarget)
{
  // Reduce the precision of the output
  
  thepitch = map(thepitch, 0, 320, -10, 10); //centre value is about 163
  theroll  = map(theroll, 0, 320, -10, 10); //centre value is about 163
  theyaw   = map(theyaw, 45, 365, -5, 5); //centre value is about 205 

  thepitch = constrain(thepitch, -10, 10);
  theroll  = constrain(theroll, -10 ,10);
  theyaw   = constrain(theyaw, -5, 5);

  // Convert to degrees
  PitchTarget = thepitch * 9;
  RollTarget  = theroll * 9;
  YawTarget   = theyaw * 9;

  PitchTarget = PitchTarget + 180;
  RollTarget  = RollTarget  + 180;
  YawTarget   = YawTarget   + 180;
}

//*********************************************************

void MoveSteppers(long PitchTarget, long RollTarget, long YawTarget)
{
 /*

  In this function we convert pitch, roll, yaw into the movements of the stepper motors.
  
     1
     |
   4-O-2 
     |
     3
  */

 long stepmotor1 = -360;
 long stepmotor2 = 0;
 long stepmotor3 = 0;
 long stepmotor4 = -360;

  /*PITCH */
  if(PitchTarget > 0)
  {
    stepmotor1 = stepmotor1 + PitchTarget;
    stepmotor3 = stepmotor3 - PitchTarget;
 
  }
  else
  if(PitchTarget < 0)
  {
    stepmotor1 = stepmotor1 - PitchTarget;
    stepmotor3 = stepmotor1 - PitchTarget;
  }

  
   /*ROLL */
  if(RollTarget > 0)
  {
    stepmotor4 = stepmotor4 + RollTarget;
    stepmotor2 = stepmotor2 - RollTarget;
 
  }
  else
  if(RollTarget < 0)
  {
    stepmotor4 = stepmotor4 - RollTarget;
    stepmotor2 = stepmotor2 - RollTarget;
  }

   /*Yaw */
  if(YawTarget > 0)
  {
    stepmotor1 = stepmotor1 + YawTarget;
    stepmotor2 = stepmotor2 + YawTarget;
    stepmotor3 = stepmotor3 + YawTarget;
    stepmotor4 = stepmotor4 + YawTarget;
 
  }
  else
  if(YawTarget < 0)
  {
    stepmotor1 = stepmotor1 - YawTarget;
    stepmotor2 = stepmotor2 - YawTarget;
    stepmotor3 = stepmotor3 - YawTarget;
    stepmotor4 = stepmotor4 - YawTarget;
  }

  stepmotor1 = stepmotor1 * degperstep;
  stepmotor2 = stepmotor2 * degperstep;
  stepmotor3 = stepmotor3 * degperstep;
  stepmotor4 = stepmotor4 * degperstep;
  
  // Time to move the stepper motors
  stepper1.moveTo(stepmotor1);
  stepper2.moveTo(stepmotor2);
  stepper3.moveTo(stepmotor3);
  stepper4.moveTo(stepmotor4);
}

//*********************************************************

void DisplayStepCurPos()
{
  Serial.print(" Stepper 1 : "); Serial.print(stepper1.currentPosition()); Serial.print(" -> ");Serial.println(stepper1.targetPosition());
  Serial.print(" Stepper 2 : "); Serial.print(stepper2.currentPosition()); Serial.print(" -> ");Serial.println(stepper2.targetPosition());
  Serial.print(" Stepper 3 : "); Serial.print(stepper3.currentPosition()); Serial.print(" -> ");Serial.println(stepper3.targetPosition());
  Serial.print(" Stepper 4 : "); Serial.print(stepper4.currentPosition()); Serial.print(" -> ");Serial.println(stepper4.targetPosition());
  Serial.println("*******************");  
}

//*********************************************************

void DisplayPitchRollYaw()
{
  Serial.print(" Pitch : "); Serial.println(pitch);
  Serial.print(" Roll  : "); Serial.println(roll);
  Serial.print(" Yaw   : "); Serial.println(yaw);
  Serial.println("*******************");
  Serial.print(" Target Pitch : "); Serial.println(PitchTarget);
  Serial.print(" Target Roll  : "); Serial.println(RollTarget);
  Serial.print(" Target Yaw   : "); Serial.println(YawTarget);
  Serial.println("*******************");  
}

//*********************************************************

void VoltageOnPin(int pwmPin, int voltage)
{
  int val;
  val = 255 * (voltage / 5);
  analogWrite(pwmPin, val);
}

//*********************************************************

void StepperMotorTest()
{
  int gotopos = 0;

  if(stepper1.currentPosition() == 180 || stepper1.currentPosition() == 0)
  {
    if (stepper1.currentPosition() == 0)
    {
      gotopos = 180;
    }
    else
    {
      gotopos = 0;
    }

    stepper1.moveTo(gotopos);
    stepper2.moveTo(gotopos);
    stepper3.moveTo(gotopos);
    stepper4.moveTo(gotopos); 
  }
}

//*********************************************************

void loop() 
{

  if (true == true)
  {
  ReadJoystick(pitch, roll, yaw);
  if( JoystickPositionChange(pitch, roll, yaw, PitchPre, RollPre, YawPre))
  {
    Joystick2StepperTargets(pitch, roll, yaw, PitchTarget, RollTarget, YawTarget);
    MoveSteppers(PitchTarget, RollTarget, YawTarget); //Only update positions when needed. Appears there is bug in lib.
  };
  
  if(debugging)
  {
    displaycounter += refreshms;
    if(displaycounter >= displayms)
    {
      // Time to show status
      DisplayStepCurPos();
      DisplayPitchRollYaw();
      displaycounter = 0;
    }
  }
  }
  //else
  //  StepperMotorTest();

  //Move Steppers if needed.
  stepper1.run();
  stepper2.run();
  stepper3.run();
  stepper4.run(); 
  
  delay(refreshms);
}
